﻿using System;
using UnityEngine;

public class InputManager : MonoBehaviour {
    private static InputManager instance;

    public static InputManager Instance
    {
        get
        {
            if (instance == null)
            {
                instance = FindObjectOfType(typeof(InputManager))
                    as InputManager;
            }

            return instance;
        }
    }

    private bool IsButtonDownPrev { get; set; }
    private bool IsButtonDown
    {
        get { return Input.GetKey(KeyCode.Space) || Input.GetKey(KeyCode.Joystick1Button0); }
    }
    
    private MotionState PrevMotionState { get; set; }
    public MotionState CurrentMotionState { get; set; }

    public ButtonStatus ButtonStat
    {
        get
        {
            if (!IsButtonDown)
                return ButtonStatus.Released;
            if (IsButtonDownPrev)
                return ButtonStatus.Holding;
            return ButtonStatus.Pressed;
        }
    }

    public InputStatus ShortButtonStat
    {
        get
        {
            if (!IsButtonDownPrev &&
                IsButtonDown)
                return InputStatus.Entered;
            return InputStatus.None;
        }
    }

    public InputStatus LongButtonStat
    {
        get
        {
            if (!IsButtonDownPrev &&
                IsButtonDown)
                return InputStatus.Entered;
            if (IsButtonDownPrev &&
                IsButtonDown)
                return InputStatus.Continuing;
            if (IsButtonDownPrev &&
                !IsButtonDown)
                return InputStatus.Stopped;
            return InputStatus.None;
        }
    }

    public InputStatus ShortMotionStat
    {
        get
        {
            if (ShortMotionToInput("Clap")            == InputStatus.Entered ||
                ShortMotionToInput("Jump")            == InputStatus.Entered ||
                ShortMotionToInput("PushUpLeft")      == InputStatus.Entered ||
                ShortMotionToInput("PushUpRight")     == InputStatus.Entered ||
                ShortMotionToInput("PushUpBoth")      == InputStatus.Entered ||
                ShortMotionToInput("JumpPushUpLeft")  == InputStatus.Entered ||
                ShortMotionToInput("JumpPushUpRight") == InputStatus.Entered ||
                ShortMotionToInput("JumpPushUpBoth")  == InputStatus.Entered ||
                ShortMotionToInput("GuardLeft")       == InputStatus.Entered ||
                ShortMotionToInput("GuardRight")      == InputStatus.Entered ||
                ShortMotionToInput("GuardBoth")       == InputStatus.Entered)
                return InputStatus.Entered;
            return InputStatus.None;
        }
    }

    public InputStatus LongMotionStat
    {
        get
        {
            if (LongMotionToInput("HandUpLeft")     == InputStatus.Entered    ||
                LongMotionToInput("HandUpLeft")     == InputStatus.Continuing ||
                LongMotionToInput("HandUpRight")    == InputStatus.Entered    ||
                LongMotionToInput("HandUpRight")    == InputStatus.Continuing ||
                LongMotionToInput("HandUpBoth")     == InputStatus.Entered    ||
                LongMotionToInput("HandUpBoth")     == InputStatus.Continuing ||
                LongMotionToInput("HandDownLeft")   == InputStatus.Entered    ||
                LongMotionToInput("HandDownLeft")   == InputStatus.Continuing ||
                LongMotionToInput("HandDownRight")  == InputStatus.Entered    ||
                LongMotionToInput("HandDownRight")  == InputStatus.Continuing ||
                LongMotionToInput("HandDownBoth")   == InputStatus.Entered    ||
                LongMotionToInput("HandDownBoth")   == InputStatus.Continuing ||
                LongMotionToInput("Jesus")          == InputStatus.Entered    ||
                LongMotionToInput("Jesus")          == InputStatus.Continuing ||
                LongMotionToInput("HeadphoneLeft")  == InputStatus.Entered    ||
                LongMotionToInput("HeadphoneLeft")  == InputStatus.Continuing ||
                LongMotionToInput("HeadphoneRight") == InputStatus.Entered    ||
                LongMotionToInput("HeadphoneRight") == InputStatus.Continuing ||
                LongMotionToInput("HeadphoneBoth")  == InputStatus.Entered    ||
                LongMotionToInput("HeadphoneBoth")  == InputStatus.Continuing ||
                LongMotionToInput("OnTheTable")     == InputStatus.Entered    ||
                LongMotionToInput("OnTheTable")     == InputStatus.Continuing)
                return InputStatus.Entered;
            return InputStatus.None;
        }
    }

    // Update is called once per frame
    void Update()
    {
        IsButtonDownPrev = IsButtonDown;
        PrevMotionState = CurrentMotionState;
    }

    public InputStatus LongMotionToInput(string name)
    {
        if (name.Contains("Both"))
        {
            var tmpname = name.Substring(name.IndexOf("Both"));

            var left  = LongMotionToInput(tmpname + "Left");
            var right = LongMotionToInput(tmpname + "Right");

            return left == right ? left : InputStatus.None;
        }

        Func<MotionState, MotionStatus> extractor = MotionStatusExtractor.GetExtractor(name);
        if (extractor == null)
            return InputStatus.None;

        MotionStatus prev = extractor(PrevMotionState),
                     curr = extractor(CurrentMotionState);

        if (prev == MotionStatus.None &&
            (curr & MotionStatus.Prepared) != MotionStatus.None)
            return InputStatus.Entered;
        if ((prev & MotionStatus.Done) != MotionStatus.None &&
            curr == MotionStatus.None)
            return InputStatus.Stopped;
        if (prev == curr &&
            (prev & (MotionStatus.Prepared | MotionStatus.Done)) != MotionStatus.None)
            return InputStatus.Continuing;

        return InputStatus.None;
    }

    public InputStatus ShortMotionToInput(string name)
    {
        if (name.Contains("Both"))
        {
            var tmpname = name.Substring(name.IndexOf("Both"));

            var left  = ShortMotionToInput(tmpname + "Left");
            var right = ShortMotionToInput(tmpname + "Right");

            return left == right ? left : InputStatus.None;
        }

        Func<MotionState, MotionStatus> extractor = MotionStatusExtractor.GetExtractor(name);
        if (extractor == null)
            return InputStatus.None;

        MotionStatus prev = extractor(PrevMotionState),
                     curr = extractor(CurrentMotionState);

        if (prev == MotionStatus.Prepared &&
            curr == MotionStatus.Done)
            return InputStatus.Entered;

        return InputStatus.None;
    }

    private static class MotionStatusExtractor
    {
        private static bool IsContain(MotionState target, MotionState motion)
        {
            return (target & motion) == motion;
        }

        // SHORT NOTES

        private static MotionStatus ExtractClap(MotionState state)
        {
            if (IsContain(state, MotionState.CLAP_PREPARE))
                return MotionStatus.Prepared;
            if (IsContain(state, MotionState.CLAP_DONE))
                return MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractJump(MotionState state)
        {
            if (IsContain(state, MotionState.JUMP_PREPARE))
                return MotionStatus.Prepared;
            if (IsContain(state, MotionState.JUMP_DONE))
                return MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractPushUpLeft(MotionState state)
        {
            if (IsContain(state, MotionState.HURRAY) &&
                IsContain(state, MotionState.HAND_MOVE_UP_LEFT))
                return MotionStatus.Prepared;
            if (IsContain(state, MotionState.HURRAY) &&
                IsContain(state, MotionState.HAND_MOVE_DOWN_LEFT))
                return MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractPushUpRight(MotionState state)
        {
            if (IsContain(state, MotionState.HURRAY) &&
                IsContain(state, MotionState.HAND_MOVE_UP_RIGHT))
                return MotionStatus.Prepared;
            if (IsContain(state, MotionState.HURRAY) &&
                IsContain(state, MotionState.HAND_MOVE_DOWN_RIGHT))
                return MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractJumpPushUpLeft(MotionState state)
        {
            if (ExtractJump      (state) == MotionStatus.Prepared &&
                ExtractPushUpLeft(state) == MotionStatus.Prepared)
                return MotionStatus.Prepared;
            if (ExtractJump      (state) == MotionStatus.Done &&
                ExtractPushUpLeft(state) == MotionStatus.Done)
                return MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractJumpPushUpRight(MotionState state)
        {
            if (ExtractJump       (state) == MotionStatus.Prepared &&
                ExtractPushUpRight(state) == MotionStatus.Prepared)
                return MotionStatus.Prepared;
            if (ExtractJump       (state) == MotionStatus.Done &&
                ExtractPushUpRight(state) == MotionStatus.Done)
                return MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractGuardLeft(MotionState state)
        {
            if (IsContain(state, MotionState.GUARD_BASE_LEFT) &&
                IsContain(state, MotionState.HAND_MOVE_UP_LEFT))
                return MotionStatus.Prepared;
            if (IsContain(state, MotionState.GUARD_BASE_LEFT) &&
                IsContain(state, MotionState.HAND_MOVE_DOWN_LEFT))
                return MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractGuardRight(MotionState state)
        {
            if (IsContain(state, MotionState.GUARD_BASE_RIGHT) &&
                IsContain(state, MotionState.HAND_MOVE_UP_RIGHT))
                return MotionStatus.Prepared;
            if (IsContain(state, MotionState.GUARD_BASE_RIGHT) &&
                IsContain(state, MotionState.HAND_MOVE_DOWN_RIGHT))
                return MotionStatus.Done;
            return MotionStatus.None;
        }

        // LONG NOTES

        private static MotionStatus ExtractHandUpLeft(MotionState state)
        {
            if (IsContain(state, MotionState.HAND_UP_LEFT))
                return MotionStatus.Prepared |
                       MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractHandUpRight(MotionState state)
        {
            if (IsContain(state, MotionState.HAND_UP_RIGHT))
                return MotionStatus.Prepared |
                       MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractHandDownLeft(MotionState state)
        {
            if (IsContain(state, MotionState.HAND_DOWN_LEFT))
                return MotionStatus.Prepared |
                       MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractHandDownRight(MotionState state)
        {
            if (IsContain(state, MotionState.HAND_DOWN_RIGHT))
                return MotionStatus.Prepared |
                       MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractJesus(MotionState state)
        {
            if (IsContain(state, MotionState.JESUS))
                return MotionStatus.Prepared |
                       MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractHeadphoneLeft(MotionState state)
        {
            if (IsContain(state, MotionState.HEADPHONE_LEFT))
                return MotionStatus.Prepared |
                       MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractHeadphoneRight(MotionState state)
        {
            if (IsContain(state, MotionState.HEADPHONE_RIGHT))
                return MotionStatus.Prepared |
                       MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractOnTheTable(MotionState state)
        {
            if (IsContain(state, MotionState.ON_THE_TABLE))
                return MotionStatus.Prepared |
                       MotionStatus.Done;
            return MotionStatus.None;
        }

        public static Func<MotionState, MotionStatus> GetExtractor(string name)
        {
            if (name.Equals("Clap"))
                return ExtractClap;
            if (name.Equals("Jump"))
                return ExtractJump;
            if (name.Equals("PushUpLeft"))
                return ExtractPushUpLeft;
            if (name.Equals("PushUpRight"))
                return ExtractPushUpRight;
            if (name.Equals("JumpPushUpLeft"))
                return ExtractJumpPushUpLeft;
            if (name.Equals("JumpPushUpRight"))
                return ExtractJumpPushUpRight;
            if (name.Equals("GuardLeft"))
                return ExtractGuardLeft;
            if (name.Equals("GuardRight"))
                return ExtractGuardRight;
            if (name.Equals("HandUpLeft"))
                return ExtractHandUpLeft;
            if (name.Equals("HandUpRight"))
                return ExtractHandUpRight;
            if (name.Equals("HandDownLeft"))
                return ExtractHandDownLeft;
            if (name.Equals("HandDownRight"))
                return ExtractHandDownRight;
            if (name.Equals("Jesus"))
                return ExtractJesus;
            if (name.Equals("HeadphoneLeft"))
                return ExtractHeadphoneLeft;
            if (name.Equals("HeadphoneRight"))
                return ExtractHeadphoneRight;
            if (name.Equals("OnTheTable"))
                return ExtractOnTheTable;
            return null;
        }
    }
}

public enum ButtonStatus
{
    Pressed,
    Released,
    Holding
}

[Flags]
public enum MotionStatus : uint
{
    Prepared = 0x1,
    Done = 0x2,
    None = 0x0
}

public enum InputStatus
{
    Entered,
    Continuing,
    Stopped,
    None
}