﻿using System;
using System.Collections.Generic;
using UnityEngine;
using Windows.Kinect;
using System.Linq;

public class MotionView : MonoBehaviour
{
	IEnumerable<KinectModule.IBody> body;
    MotionDiscriminator discriminator = new MotionDiscriminator();

    // Use this for initialization
    void Start ()
    {
        body = GameObject.Find("Managers").GetComponent<CoordinateMapperManager>().BodyBuffer;

        discriminator.Preset(body.ToList()[0]);
	}
	
	// Update is called once per frame
	void Update ()
    {
        InputManager.Instance.CurrentMotionState = DetermineState();
	}

	MotionState DetermineState()
	{
        var tracked = body.Where(x => x != null && x.IsTracked).ToList();
        if (tracked.Count() == 0 || tracked[0].Joints == null)
            return MotionState.UNKNOWN;

        discriminator.Update(tracked[0]);

        return discriminator.Motion;
	}
}

[Flags]
public enum MotionState : uint
{
    UNKNOWN              = 0x00000,
    CLAP_PREPARE         = 0x00001,
    CLAP_DONE            = 0x00002,
	JUMP_PREPARE         = 0x00004,
	JUMP_DONE            = 0x00008,
    HURRAY               = 0x00010,
	HAND_MOVE_UP_LEFT    = 0x00020,
	HAND_MOVE_DOWN_LEFT  = 0x00040,
	HAND_MOVE_UP_RIGHT   = 0x00080,
	HAND_MOVE_DOWN_RIGHT = 0x00100,
	GUARD_BASE_LEFT      = 0x00200,
	GUARD_BASE_RIGHT     = 0x00400,
	HAND_UP_LEFT         = 0x00800,
	HAND_DOWN_LEFT       = 0x01000,
	HAND_UP_RIGHT        = 0x02000,
	HAND_DOWN_RIGHT      = 0x04000,
	JESUS                = 0x10000,
	HEADPHONE_LEFT       = 0x20000,
	HEADPHONE_RIGHT      = 0x40000,
	ON_THE_TABLE         = 0x80000
}

internal class MotionDiscriminator
{
    DistItvExtractor extractor;

    private MotionState _Motion;

    public MotionState Motion
    { get { return _Motion; } }

    public void Preset(KinectModule.IBody body)
    {
        CameraSpacePoint KneeLeftBase  = body.Joints[JointType.KneeLeft ].Position,
                         KneeRightBase = body.Joints[JointType.KneeRight].Position;

        extractor = new DistItvExtractor((KneeLeftBase.Y + KneeRightBase.Y) / 2);

        Initialize();
    }

    private void Initialize()
    {
        _Motion = MotionState.UNKNOWN;
    }

    public void Update(KinectModule.IBody body)
    {
        extractor.Extract(body);

        Initialize();
        Determine();
    }

    float Distance(CameraSpacePoint a, CameraSpacePoint b)
    {
        return Mathf.Sqrt(Mathf.Pow(a.X - b.X, 2.0f) +
                          Mathf.Pow(a.Y - b.Y, 2.0f) +
                          Mathf.Pow(a.Z - b.Z, 2.0f));
    }

    void Determine()
    {
        Clap();
        Jump();
        Hurray();
        HandMove();
        GuardBase();
        HandUp();
        HandDown();
        Jesus();
        Headphone();
        OnTheTable();
    }

    void Clap()
    {
        const float distPrepare = 0.3f, distDone = 0.1f;

        if (extractor.DistHandBaseSpineShoulder <= 0.0f)
            return;

        if      (extractor.DistHand > distPrepare)
            _Motion |= MotionState.CLAP_PREPARE;
        else if (extractor.DistHand < distDone)
            _Motion |= MotionState.CLAP_DONE;
    }

    void Jump()
    {
        const float distPrepare = 0.05f, distDone = 0.0f;
        
        if      (extractor.DistSpine < distPrepare)
            _Motion |= MotionState.JUMP_PREPARE;
        else if (extractor.DistSpine > distDone)
            _Motion |= MotionState.JUMP_DONE;
    }

    void Hurray()
    {
        if (extractor.DistHandBaseHead > 0)
            _Motion |= MotionState.HURRAY;
    }

    void HandMove()
    {
        if      (extractor.DistHandLeft > 0)
            _Motion |= MotionState.HAND_MOVE_UP_LEFT;
        else if (extractor.DistHandLeft < 0)
            _Motion |= MotionState.HAND_MOVE_DOWN_LEFT;

        if      (extractor.DistHandRight > 0)
            _Motion |= MotionState.HAND_MOVE_UP_RIGHT;
        else if (extractor.DistHandRight < 0)
            _Motion |= MotionState.HAND_MOVE_DOWN_RIGHT;
    }

    void GuardBase()
    {
        const float distBase = 0.5f;

        if (extractor.DistHandBaseElbow_Left > 0.0f &&
            extractor.ItvElbowBaseSpineMid_Left < distBase)
            _Motion |= MotionState.GUARD_BASE_LEFT;

        if (extractor.DistHandBaseElbow_Right > 0.0f &&
            extractor.ItvElbowBaseSpineMid_Right < distBase)
            _Motion |= MotionState.GUARD_BASE_RIGHT;
    }

    void HandUp()
    {
        const float distUp = 0.2f;

        if (extractor.DistHandBaseHead_Left > distUp)
            _Motion |= MotionState.HAND_UP_LEFT;

        if (extractor.DistHandBaseHead_Right > distUp)
            _Motion |= MotionState.HAND_UP_RIGHT;
    }

    void HandDown()
    {   
        if (extractor.DistHandBaseSpineMid_Left < 0.0f)
            _Motion |= MotionState.HAND_DOWN_LEFT;

        if (extractor.DistHandBaseSpineMid_Right < 0.0f)
            _Motion |= MotionState.HAND_DOWN_RIGHT;
    }

    void Jesus()
    {
        const float distJesus = 0.5f;

        if (extractor.DistHandBaseSpineShoulder > 0.0f && extractor.ItvHandBaseHead >= distJesus)
            _Motion |= MotionState.JESUS;
    }

    void Headphone()
    {
        const float itvDepth = 0.2f, itvEar = 0.2f;

        if (extractor.ItvHandBaseHead_LeftEar < itvEar &&
            extractor.ItvHandBaseHead_LeftDepth < itvDepth)
            _Motion |= MotionState.HEADPHONE_LEFT;

        if (extractor.ItvHandBaseHead_RightEar < itvEar &&
            extractor.ItvHandBaseHead_RightDepth < itvDepth)
            _Motion |= MotionState.HEADPHONE_RIGHT;
    }

    void OnTheTable()
    {
        const float distTable = 0.5f;

        if (extractor.DistKneeBaseTable >= distTable)
            _Motion |= MotionState.ON_THE_TABLE;
    }
}

internal class DistItvExtractor
{
    CameraSpacePoint Head,
                     HandLeft,
                     HandRight,
                     SpineShoulder,
                     SpineMid,
                     ElbowLeft,
                     ElbowRight,
                     KneeLeft,
                     KneeRight,
                     SpineMidRecent,
                     HandLeftRecent,
                     HandRightRecent;

    float Table;

    public float DistHandBaseSpineShoulder{ get; private set; }
    public float DistHand{ get; private set; }
    public float DistSpine{ get; private set; }
    public float DistHandBaseHead{ get; private set; }
    public float DistHandLeft{ get; private set; }
    public float DistHandRight{ get; private set; }
    public float DistHandBaseElbow_Left{ get; private set; }
    public float DistHandBaseElbow_Right{ get; private set; }
    public float DistHandBaseHead_Left{ get; private set; }
    public float DistHandBaseHead_Right{ get; private set; }
    public float DistHandBaseSpineMid_Left{ get; private set; }
    public float DistHandBaseSpineMid_Right{ get; private set; }
    public float DistKneeBaseTable{ get; private set; }
    public float ItvElbowBaseSpineMid_Left{ get; private set; }
    public float ItvElbowBaseSpineMid_Right{ get; private set; }
    public float ItvHandBaseHead{ get; private set; }
    public float ItvHandBaseHead_LeftEar{ get; private set; }
    public float ItvHandBaseHead_RightEar{ get; private set; }
    public float ItvHandBaseHead_LeftDepth{ get; private set; }
    public float ItvHandBaseHead_RightDepth{ get; private set; }

    public DistItvExtractor(float table)
    {
        Table = table;
    }

    public void Extract(KinectModule.IBody body)
    {
        UpdatePosition(body);
        UpdateDistItv();
        SaveRecent();
    }

    void UpdatePosition(KinectModule.IBody body)
    {
        Head          = body.Joints[JointType.Head         ].Position;
        HandLeft      = body.Joints[JointType.HandLeft     ].Position;
        HandRight     = body.Joints[JointType.HandRight    ].Position;
        SpineShoulder = body.Joints[JointType.SpineShoulder].Position;
        SpineMid      = body.Joints[JointType.SpineMid     ].Position;
        ElbowLeft     = body.Joints[JointType.ElbowLeft    ].Position;
        ElbowRight    = body.Joints[JointType.ElbowRight   ].Position;
        KneeLeft      = body.Joints[JointType.KneeLeft     ].Position;
        KneeRight     = body.Joints[JointType.KneeRight    ].Position;
    }

    void SaveRecent()
    {
        SpineMidRecent  = SpineMid;
        HandLeftRecent  = HandLeft;
        HandRightRecent = HandRight;
    }

    void UpdateDistItv()
    {
        DistHandBaseSpineShoulder = Mathf.Min(HandLeft.Y, HandRight.Y) - SpineShoulder.Y;
        
        DistHand = Distance(HandLeft, HandRight);
        
        DistSpine = SpineMid.Y - SpineMidRecent.Y;
        
        DistHandBaseHead = Mathf.Min(HandLeft.Y, HandRight.Y) - Head.Y;
        
        DistHandLeft  = HandLeft.Y  - HandLeftRecent.Y;
        DistHandRight = HandRight.Y - HandRightRecent.Y;
        
        DistHandBaseElbow_Left  = HandLeft.Y  - ElbowLeft.Y;
        DistHandBaseElbow_Right = HandRight.Y - ElbowRight.Y;
        
        DistHandBaseHead_Left  = HandLeft.Y  - Head.Y;
        DistHandBaseHead_Right = HandRight.Y - Head.Y;
        
        DistHandBaseSpineMid_Left  = HandLeft.Y  - SpineMid.Y;
        DistHandBaseSpineMid_Right = HandRight.Y - SpineMid.Y;
        
        DistKneeBaseTable = Mathf.Min(KneeLeft.Y, KneeRight.Y) - Table;

        ItvElbowBaseSpineMid_Left  = Mathf.Abs(ElbowLeft.X  - SpineMid.X);
        ItvElbowBaseSpineMid_Right = Mathf.Abs(ElbowRight.X - SpineMid.X);
        
        ItvHandBaseHead = Mathf.Min(Mathf.Abs(HandLeft.X - Head.X), Mathf.Abs(HandLeft.X - Head.X));
        
        ItvHandBaseHead_LeftEar    = Mathf.Abs(HandLeft.X  - Head.X);
        ItvHandBaseHead_RightEar   = Mathf.Abs(HandRight.X - Head.X);
        ItvHandBaseHead_LeftDepth  = Mathf.Abs(HandLeft.Z  - Head.Z);
        ItvHandBaseHead_RightDepth = Mathf.Abs(HandRight.Z - Head.Z);
    }

    float Distance(CameraSpacePoint a, CameraSpacePoint b)
    {
        return Mathf.Sqrt(Mathf.Pow(a.X - b.X, 2.0f) +
                          Mathf.Pow(a.Y - b.Y, 2.0f) +
                          Mathf.Pow(a.Z - b.Z, 2.0f));
    }
}